;;; -*- Mode: LISP; Syntax: ANSI-Common-Lisp; Base: 10 -*-
;; See the file LICENCE for licence information.

;; CL-STORE now has a concept of backends.
;; store and restore now take an optional  backend as an
;;  argument to do the actual restoring. Examples of use are 
;; in default-backend.lisp and xml-backend.lisp

(in-package :cl-store)
(declaim (optimize (speed 3) (safety 0) (debug 0)))


(defun required-arg (name)
  (error "~A is a required argument" name))

(defclass backend ()
  ((name :accessor name :initform "Unknown" :initarg :name)
   (magic-number :accessor magic-number :initarg :magic-number)
   (stream-type :accessor stream-type :initarg :stream-type
                :initform (required-arg "stream-type"))
   (restorer-funs :accessor restorer-funs :initform (make-hash-table)
                  :initarg :restorer-funs))
  (:documentation "Core class which custom backends must extend"))

(defparameter *registered-backends* nil 
  "An assoc list mapping backend-names to the backend objects")

(defun mkstr (&rest args)
  (with-output-to-string (s)
    (dolist (x args)
      (princ x s))))

(defun symbolicate (&rest syms)
  "Concatenate all symbol names into one big symbol"
  (values (intern (apply #'mkstr syms))))

(defun get-store-macro (name class-name)
  "Return the defstore-? macro which will be used by a custom backend"
  (let ((macro-name (symbolicate 'defstore- name)))
    `(defmacro ,macro-name ((var type stream &key qualifier) 
                            &body body)
      `(defmethod internal-store-object ,@(if qualifier (list qualifier) nil)
        ((,var ,type) ,stream (backend ,',class-name))
        ,(format nil "Definition for storing an object of type ~A with ~
backend ~A" type ',name)
        ,@body))))

(defun get-restore-macro (name)
  "Return the defrestore-? macros which will be used by a custom backend"
  (let ((macro-name (symbolicate 'defrestore- name)))
    `(defmacro ,macro-name ((type place) &body body)
      (let ((fn-name (gensym (symbol-name (symbolicate ',name '- type)))))
        `(flet ((,fn-name (,place) ,@body))
          (let* ((backend (find-backend ',',name))
                 (restorers (restorer-funs backend)))
            (when (gethash ',type restorers)
              (warn "Redefining restorer ~A for backend ~(~A~)"
                    ',type (name backend)))
            (setf (gethash ',type restorers)
                  #',fn-name)))))))

(defun real-stream-type (value)
  (ecase value
    (char 'character)
    (binary 'integer)))

(defun register-backend (name class magic-number stream-type)
  (declare (type symbol name))
  (assert (member stream-type '(char binary)))
  (let ((instance (make-instance class
                                 :name name
                                 :magic-number magic-number
                                 :stream-type (real-stream-type stream-type))))
    (if (assoc name *registered-backends*)
        (cerror "Redefine backend" "Backend is already defined ~A" name)
        (push (cons name instance) *registered-backends*))
    instance))

(defun find-backend (name)
  (declare (type symbol name))
  "Return backup called NAME or NIL if not found."
  (cdr (assoc name *registered-backends*)))

(defun get-class-form (name fields extends)
  `(defclass ,name (,extends)
    ,fields
    (:documentation ,(format nil "Autogenerated cl-store class for backend ~(~A~)." 
                             name))))

(defmacro defbackend (name &key (stream-type (required-arg "stream-type"))
                      (magic-number nil) fields (extends 'backend))
  "Defines a new backend called NAME. Stream type must be either 'char or 'binary. 
FIELDS is a list of legal slots for defclass. MAGIC-NUMBER, when supplied, will
be written down stream as verification and checked on restoration.
EXTENDS is a class to extend, which must be backend or a class which extends
backend"
  (assert (symbolp name))
  (let ((class-name (symbolicate name '-backend)))
    `(eval-when (:compile-toplevel :load-toplevel :execute)
      (prog2 
          ,(get-class-form class-name fields extends)
          (register-backend ',name ',class-name ,magic-number ,stream-type )
        ,(get-store-macro name class-name)
        ,(get-restore-macro name)))))


(defmacro with-backend ((backend-name) &body body)
  "Run BODY with *default-backend* bound to the backend BACKEND-NAME"
  `(let ((*default-backend* (or (find-backend ',backend-name)
                                (error "Can't find backend ~A" 
                                       ',backend-name))))
    ,@body))

;; EOF