;;; -*- Mode: LISP; Syntax: ANSI-Common-Lisp; Base: 10 -*-
;; See the file LICENCE for licence information.

;; Defines a special backend type which specializes various methods
;; in plumbing.lisp to make it nice and easy to 
;; resolve possible circularities in objects.
;; Most of the work is done using the resolving-object
;; macro which knows how to handle an object which 
;; is a referrer to a previously restored value.
;; Backends wanting to make use of this should take
;; a look at default-backend.lisp and xml-backend.lisp
;; paying special attention to the defbackend form and the 
;; defrestore definitions for cons, array, simple-vector
;; array and hash-table.
;;
;; As a note this will ignore integers, symbols or characters
;; as referrer values. It will handle all other EQ number although
;; software depending on eq numbers are not conforming 
;; programs according to the Hyperspec(notes in EQ).

(in-package :cl-store)
(declaim (optimize (speed 3) (safety 0) (debug 0)))

(defvar *referrer-string* "%%Referrer-" 
  "String which will be interned to create a symbol we
  can recognize as a referrer.")

(defvar *prefix-setters* 
  '(slot-value aref row-major-aref)
  "Setfable places which take the object to set before the 
  rest of the arguments.")

(defun get-setf-place  (place obj)
  "Return a legal setf form for setting PLACE in OBJ, see *prefix-setters*."
  (declare (type (or cons symbol) place))
  (cond ((atom place) `(,place ,obj))
        ((member (car place) *prefix-setters*)
         `(,(car place) ,obj ,@(cdr place)))
        (t `(,@place ,obj))))


;; The definitions for setting and setting-hash sits in resolving-object.
(defmacro setting (place get)
  "Resolve the possible referring object retrieved by GET and 
  set it into PLACE. Only usable within a resolving-object form."
  (declare (ignore place get))
  (error "setting can only be used inside a resolving-object form."))

(defmacro setting-hash (getting-key getting-value)
  "Insert the value retrieved by GETTING-VALUE with the key 
  retrieved by GETTING-KEY, resolving possible circularities.
  Only usable within a resolving-object form."
  (declare (ignore getting-key getting-value))
  (error "setting-hash can only be used inside a resolving-object form."))


(defmacro resolving-object (create &body body)
  "Execute body attempting to resolve circularities found in 
   form CREATE."
  (with-gensyms (obj value key)
    `(macrolet ((setting (place getting)
                 (let ((setf-place (get-setf-place place ',obj)))
                   `(let ((,',value ,getting))
                     (if (referrerp ,',value)
                         (push (lambda ()
                                 (setf ,setf-place
                                       (referred-value ,',value 
                                                       *restored-values*)))
                               *need-to-fix*)
                         (setf ,setf-place ,',value)))))
                (setting-hash (getting-key getting-place)
                 `(let ((,',key ,getting-key))
                   (if (referrerp ,',key)
                       (let ((,',value ,getting-place))
                         (push (lambda () 
                                 (setf (gethash 
                                        (referred-value ,',key *restored-values*)
                                        ,',obj)
                                       (if (referrerp ,',value)
                                           (referred-value ,',value 
                                                           *restored-values*)
                                           ,',value)))
                               *need-to-fix*))
                       (setting (gethash ,',key) ,getting-place)))))
      (let ((,obj ,create))
        ,@body
        ,obj))))

(defun referrerp (val)
  "Is val a referrer?"
  (and (symbolp val)
       (eq (symbol-package val) #.(find-package :cl-store-referrers))
       (equal (subseq (symbol-name val) 0 11)
              *referrer-string*)))

(defun referred-value (referrer hash)
  "Return the value REFERRER is meant to be by looking in HASH."
  (gethash (read-from-string (subseq (symbol-name referrer) 11))
           hash))


(defun make-referrer (x)
  "Create a new referrer suffixed with X."
  (declare (type fixnum x))
  (let ((name (intern (format nil "%%Referrer-~D" x)
                      :cl-store-referrers)))
    name))


(defclass resolving-backend (backend) 
  ()
  (:documentation "A backend which does the setup for resolving circularities."))

(declaim (type fixnum *stored-counter*))
(defvar *stored-counter*)
(defvar *stored-values*)


(defmethod backend-store ((obj t) (place t) (backend resolving-backend))
  "Store OBJ into PLACE. Does the setup for counters and seen values."
  (let ((*stored-counter* 0) 
        (*stored-values* (make-hash-table :test #'eq))) 
    (check-stream-element-type place backend)
    (store-backend-code place backend)
    (backend-store-object obj place backend)
    obj))

(defun seen (obj)
  "Has this object already been stored?"
  (incf *stored-counter*)
  (gethash obj *stored-values*))

(defun update-seen (obj)
  "Register OBJ as having been stored."
  (setf (gethash obj *stored-values*)  *stored-counter*)
  obj)

(deftype not-circ ()
  "Type grouping integer, characters and symbols, which we
  don't bother to check if they have been stored before"
  '(or integer character symbol))

(defun needs-checkp (obj)
  "Do we need to check if this object has been stored before?"
  (not (typep obj 'not-circ)))

(defun value-or-referrer (obj)
  "Returns the number of the referrer and t if this object 
  has already been stored in this STORE call."
  (if (needs-checkp obj)
      (aif (seen obj)
           (values it t)
           (values (update-seen obj) nil))
      obj))

(defgeneric store-referrer (obj place backend)
  (:documentation "Store the number OBJ into PLACE as a referrer for BACKEND.")
  (:method ((obj t) (place t) (backend resolving-backend))
    (store-error  "store-referrer must be specialized for backend ~(~A~)."
                  (name backend))))

(defmethod backend-store-object ((obj t) (place t) (backend resolving-backend))
  "Store object if we have not seen this object before, otherwise retrieve
  the referrer object for it and store that using store-referrer."
  (multiple-value-bind (obj referrerp) (value-or-referrer obj)
    (if referrerp
        (store-referrer obj place backend)
        (internal-store-object obj place backend))))



;; Restoration.
(declaim (type fixnum *restore-counter*))
(defvar *restore-counter*)
(defvar *need-to-fix*)
(defvar *restored-values*)

(defmethod backend-restore ((place stream) (backend resolving-backend))
  "Restore an object from PLACE using BACKEND. Does the setup for 
  various variables used by resolving-object."
  (let ((*restore-counter* 0)
        (*need-to-fix* nil)
        (*restored-values* (make-hash-table)))
    (check-stream-element-type place backend)
    (check-magic-number place backend)
    (let ((obj (backend-restore-object place backend)))
      (dolist (fn *need-to-fix*)
        (funcall (the function fn)))
      obj)))

(defmethod backend-restore-object ((place t) (backend resolving-backend))
  "Retrieve a object from PLACE, does housekeeping for circularity fixing."
  (let ((reader (find-function-for-type place backend)))
    (if (not (int-sym-or-char-p reader backend))
        (setf (gethash (incf *restore-counter*) *restored-values*)
              (new-val (funcall (the function reader) place)))
        (funcall (the function reader) place))))

(defun int-sym-or-char-p (fn backend)
  "Is function FN registered to restore an integer, character or symbol
  in BACKEND."
  (let ((readers (restorer-funs backend)))
    (or (eq fn (lookup-reader 'integer readers))
        (eq fn (lookup-reader 'character readers))
        (eq fn (lookup-reader 'symbol readers)))))
  

(defun new-val (val)
  "Tries to get a referred value to reduce unnecessary cirularity fixing."
  (if (referrerp val)
      (aif (referred-value val *restored-values*)
           it
           val)
      val))

;; EOF