;;; -*- Mode: LISP; Syntax: ANSI-Common-Lisp; Base: 10 -*-
;; See the file LICENCE for licence information.

;; The cl-store backend. 

;;  Outstanding objects.
;;  functions, methods
;;  closures (once done add initform, and default-initargs)


(in-package :cl-store)

(declaim (optimize (speed 3) (safety 0) (debug 0)))

(eval-when (:compile-toplevel :load-toplevel :execute)
  (defvar *cl-store-backend*
    (defbackend cl-store :magic-number 1886611788 :stream-type 'binary
                :extends resolving-backend
                :fields ((restorers :accessor restorers :initform
                                    nil))))
  (defun register-code (code name)
    (push (cons code name) (restorers *cl-store-backend*))
    code))

;;  Type code constants
(defconstant +referrer-code+ (register-code 1 'referrer))
(defconstant +non-return-code+ (register-code 2 'non-return))
(defconstant +integer-code+ (register-code 4 'integer))
(defconstant +simple-string-code+ (register-code 5 'simple-string))
(defconstant +float-code+ (register-code 6 'float))
(defconstant +ratio-code+ (register-code 7 'ratio))
(defconstant +character-code+ (register-code 8 'character))
(defconstant +complex-code+ (register-code 9 'complex))
(defconstant +symbol-code+ (register-code 10 'symbol))
(defconstant +cons-code+ (register-code 11 'cons))
(defconstant +pathname-code+ (register-code 12 'pathname))
(defconstant +hash-table-code+ (register-code 13 'hash-table))
(defconstant +standard-object-code+ (register-code 14 'standard-object))
(defconstant +condition-code+ (register-code 15 'condition))
(defconstant +structure-object-code+ (register-code 16 'structure-object))
(defconstant +standard-class-code+ (register-code 17 'standard-class))
(defconstant +built-in-class-code+ (register-code 18 'built-in-class))
(defconstant +array-code+ (register-code 19 'array))
(defconstant +simple-vector-code+ (register-code 20 'simple-vector))
(defconstant +package-code+ (register-code 21 'package))

;; setups for type code mapping
(defun output-type-code (code stream)
  (declare (type (mod 256) code))
  (write-byte (ldb (byte 8 0) code) stream))

(defun read-type-code (stream)
  (read-byte stream))



;; get-next-reader needs to return a symbol which will be used by the
;; backend to lookup the function that was defined by
;; defrestore-cl-store to restore it, or nil if not found. 
(defmethod get-next-reader ((stream stream) (backend cl-store-backend))
  (cdr (assoc (read-type-code stream)
              (restorers backend))))


;; referrer, Required for a resolving backend
(defmethod store-referrer (ref stream (backend cl-store-backend))
  (output-type-code +referrer-code+ stream)
  (store-32-byte ref stream))

(defrestore-cl-store (referrer stream)
  (make-referrer (read-32-byte stream nil)))



;; non return only used with standard-classes
(defun store-non-return (obj stream)
  (output-type-code +non-return-code+ stream)
  (store-object obj stream))

(defrestore-cl-store (non-return stream)
  (restore-object stream)
  (restore-object stream))
                                 

;; integers
(defstore-cl-store (obj integer stream)
  (output-type-code +integer-code+ stream)
  (loop for n = (abs obj) then (ash n -32)
        for counter from 0
        with collect = nil
        until (zerop n)
        do (push n collect)
        finally (progn
                  (store-32-byte (if (minusp obj)
                                    (- counter)
                                    counter)
                                stream)
                  (dolist (num collect)
                    (store-32-byte num stream)))))

(defrestore-cl-store (integer buff)
  (let ((count (read-32-byte buff))
        (result 0))
    (loop repeat (abs count) do
          (setf result (+ (ash result 32) (read-32-byte buff nil))))
    (if (minusp count)
        (- result)
        result)))


;; Strings
;; If the string to be stored is of type simple-standard-string
;; we can write it down byte by byte. Otherwise we treat it as 
;; an array.
(deftype simple-standard-string ()
  `(simple-array standard-char (*)))

(defun output-simple-standard-string (obj stream)
  (store-32-byte (length obj) stream)
  (dotimes (x (length obj))
    (write-byte (char-code (schar obj x)) stream)))

(defun restore-simple-standard-string (stream)
  (let* ((length (read-32-byte stream nil))
         (res (make-string length)))
    (dotimes (x length)
      (setf (schar res x) (code-char (read-byte stream))))
    res))

(defun store-simple-standard-string (string stream)
  (output-type-code +simple-string-code+ stream)
  (output-simple-standard-string string stream))


(defstore-cl-store (obj string stream)
  (if (typep obj 'simple-standard-string)
      (store-simple-standard-string obj stream)
      (store-array obj stream)))

(defrestore-cl-store (simple-string stream)
  (restore-simple-standard-string stream))

;; Floats
;; Is integer-decode-float the Right Thing, or should we 
;; be using something like sb-kernel:single-float-bits
;; and sb-kernel:make-single-float
(defstore-cl-store (obj float stream)
  (output-type-code +float-code+ stream)    
  (multiple-value-bind (significand exponent sign)
      (integer-decode-float obj)
    (write-byte (float-type obj) stream)
    (store-object significand stream)
    (store-object exponent stream)
    (store-object sign stream)))

(defrestore-cl-store (float stream)
  (let ((type (get-float-type (read-byte stream)))
        (significand (restore-object stream))
        (exponent (restore-object stream))
        (sign (restore-object stream)))
    (float (* (* significand (* 1.0d0 (expt 2 exponent))) sign) type)))

;; ratio
(defstore-cl-store (obj ratio stream)
  (output-type-code +ratio-code+ stream)
  (store-object (numerator obj) stream)
  (store-object (denominator obj) stream))

(defrestore-cl-store (ratio stream)
  (/ (restore-object stream) (restore-object stream)))

;; chars
(defstore-cl-store (obj character stream)
  (output-type-code +character-code+ stream)    
  (store-object (char-code obj) stream))


(defrestore-cl-store (character stream)
  (code-char (restore-object stream)))

;; complex
(defstore-cl-store (obj complex stream)
  (output-type-code +complex-code+ stream)    
  (let ((real (realpart obj))
        (imag (imagpart obj)))
    (store-object real stream)
    (store-object imag stream)))

(defrestore-cl-store (complex stream)
  (let ((real (restore-object stream))
        (imag (restore-object stream)))
    (complex real imag)))

;; symbols
(defstore-cl-store (obj symbol stream)
  (output-type-code +symbol-code+ stream)    
  (output-simple-standard-string (package-name (or (symbol-package obj)
                                                   *package*))
                                 stream)
  (output-simple-standard-string (symbol-name obj)
                                 stream))

(defrestore-cl-store (symbol stream)
  (let ((package (restore-simple-standard-string stream))
        (name (restore-simple-standard-string stream)))
    (multiple-value-bind (a b)
        (intern name package)
      (declare (ignore b))
      a)))

;; lists
(defstore-cl-store (obj cons stream)
  (output-type-code +cons-code+ stream)    
  (store-object (car obj) stream)
  (store-object (cdr obj) stream))

;; this is an examples of a restorer which handles 
;; circularities using resolving-object and setting.
(defrestore-cl-store (cons stream)
  (resolving-object (cons nil nil)
    (setting car (restore-object stream))
    (setting cdr (restore-object stream))))

;; pathnames
(defstore-cl-store (obj pathname stream)
  (output-type-code +pathname-code+ stream)
  (store-object (pathname-device obj) stream)
  (store-object (pathname-directory obj) stream)
  (store-object (pathname-name obj) stream)
  (store-object (pathname-type obj) stream)
  (store-object (pathname-version obj) stream))

(defrestore-cl-store (pathname stream)
  (make-pathname    
   :device (restore-object stream)
   :directory (restore-object stream)
   :name (restore-object stream)
   :type (restore-object stream)
   :version (restore-object stream)))


;; hash tables
(defstore-cl-store (obj hash-table stream)
  (output-type-code +hash-table-code+ stream)    
  (store-object (hash-table-rehash-size obj) stream)
  (store-object (hash-table-rehash-threshold obj) stream)
  (store-object (hash-table-size obj) stream)
  (store-object (hash-table-test obj) stream)
  (store-object (hash-table-count obj) stream)
  (loop for key being the hash-keys of obj
        for value being the hash-values of obj do
        (store-object key stream)
        (store-object value stream)))

(defrestore-cl-store (hash-table stream)
  (let ((rehash-size (restore-object stream))
        (rehash-threshold (restore-object stream))
        (size (restore-object stream))
        (test (restore-object stream))
        (count (restore-object stream)))
    (let ((hash (make-hash-table :test (symbol-function test)
                                 :rehash-size rehash-size
                                 :rehash-threshold rehash-threshold
                                 :size size)))
      (resolving-object hash
        (loop repeat count do
              ;; Unfortunately we can't use the normal setting here
              ;; since there could be a circularity in the key
              ;; and we need to make sure that both objects are 
              ;; removed from the stream at this point.
              (setting-hash (restore-object stream) 
                            (restore-object stream))))
      hash)))

(defun store-type-object (obj stream)
  (let* ((all-slots (remove-if-not (lambda (x)
                                     (slot-boundp obj (slot-definition-name x)))
                                   (compute-slots (class-of obj))))
         (slots (if *store-class-slots* 
                    all-slots
                    (remove-if #'(lambda (x) (eql (slot-definition-allocation x)
                                                  :class))
                               all-slots))))
    (store-object (type-of obj) stream)
    (store-object (length slots) stream)
    (dolist (slot slots)
      (let ((slot-name (slot-definition-name slot)))
        (store-object slot-name stream)
        (store-object (slot-value obj slot-name) stream)))))


(defstore-cl-store (obj standard-object stream)
  (output-type-code +standard-object-code+ stream)    
  (store-type-object obj stream))

(defstore-cl-store (obj condition stream)
  (output-type-code +condition-code+ stream)    
  (store-type-object obj stream))

#+(or sbcl cmu)
(defstore-cl-store (obj structure-object stream)
  (output-type-code +structure-object-code+ stream)
  (store-type-object obj stream))

(defun restore-type-object (stream)
  (let* ((class (find-class (restore-object stream)))
         (length (restore-object stream))
         (new-instance (allocate-instance class)))
    (loop repeat length do
          (let ((slot-name (restore-object stream)))
            ;; slot-names are always symbols so we don't
            ;; have to worry about circularities
            (resolving-object new-instance
              (setting (slot-value slot-name) (restore-object stream)))))
    new-instance))

#+(or sbcl cmu)
(defrestore-cl-store (structure-object stream)
  (restore-type-object stream))

(defrestore-cl-store (condition stream)
  (restore-type-object stream))

(defrestore-cl-store (standard-object stream)
  (restore-type-object stream))

;; classes
(defstore-cl-store (obj standard-class stream)
  (output-type-code +standard-class-code+ stream)
  (when *store-class-superclasses*
    (loop for x in (class-direct-superclasses obj) do
          (when (and x (not (eql x #.(find-class 'standard-object))))
            (store-non-return x stream))))
  (store-object (get-class-details obj) stream))

(defrestore-cl-store (standard-class stream)
  (let* ((vals (restore-object stream))
         (keywords '(:direct-slots :direct-superclasses
                     :metaclass))
         (final (mappend #'list keywords (cdr vals)))
         (class (car vals)))
    (cond ((find-class class nil)
           (cond (*nuke-existing-classes*
                  (apply #'ensure-class class final)
                  #+clisp (add-methods-for-class class (second vals)))
                 (t (find-class class))))
          (t (apply #'ensure-class class final)
             #+clisp (add-methods-for-class class (second vals))))))

;; built in classes
(defstore-cl-store (obj built-in-class stream)
  (output-type-code +built-in-class-code+ stream)
  (store-object (class-name obj) stream))

(defmethod internal-store-object ((obj (eql (find-class 'hash-table))) stream
                                  (backend cl-store-backend))
  (output-type-code +built-in-class-code+ stream)
  (store-object 'cl:hash-table stream))

(defrestore-cl-store (built-in-class stream)
  (find-class (restore-object stream)))

;; arrays and vectors
(defstore-cl-store (obj array stream)
  (store-array obj stream))

(defun store-array (obj stream)
  (output-type-code +array-code+ stream)
  (if (and (= (array-rank obj) 1)
           (array-has-fill-pointer-p obj))
      (store-object (fill-pointer obj) stream)
      (store-object nil stream))
  (store-object (array-element-type obj) stream)
  (store-object (adjustable-array-p obj) stream)
  (store-object (array-dimensions obj) stream)
  (dolist (x (multiple-value-list (array-displacement obj)))
    (store-object x stream))
  (store-object (array-total-size obj) stream)
  (loop for x from 0 to (1- (array-total-size obj)) do
        (store-object (row-major-aref obj x) stream)))

(defrestore-cl-store (array stream)
  (let* ((fill-pointer (restore-object stream))
         (element-type (restore-object stream))
         (adjustable (restore-object stream))
         (dimensions (restore-object stream))
         (displaced-to (restore-object stream))
         (displaced-offset (restore-object stream))
         (size (restore-object stream))
         (res (make-array dimensions  
                          :element-type element-type
                          :adjustable adjustable
                          :fill-pointer fill-pointer)))
    (when displaced-to 
      (adjust-array res dimensions :displaced-to displaced-to
                    :displaced-index-offset displaced-offset))
    (resolving-object res
      (loop for x from 0 to (1- size) do
            (let ((pos x))
              (setting (row-major-aref pos) (restore-object stream)))))
    res))


;; clisp doesn't have the class simple-vector
#-clisp
(defstore-cl-store (obj simple-vector stream)
  (output-type-code +simple-vector-code+ stream)
  (let ((size (length obj)))
    (store-object size stream)
    (loop for x across obj do
          (store-object x stream))))

#-clisp
(defrestore-cl-store (simple-vector stream)
  (let* ((size (restore-object stream))
         (res (make-array size)))
    (resolving-object res
      (loop repeat size
            for i from 0 do
            ;; we need to copy the index so that
            ;; it's value is preserved for after the loop.
            (let ((x i)) 
              (setting (aref x) (restore-object stream)))))
    res))

;; packages
(defstore-cl-store (obj package stream)
  (output-type-code +package-code+ stream)
  (store-object (package-name obj) stream))

(defrestore-cl-store (package stream)
  (find-package (restore-object stream)))

(setf *default-backend* (find-backend 'cl-store))
;; EOF
