;;; -*- Mode: LISP; Syntax: ANSI-Common-Lisp; Base: 10 -*-
;; See the file LICENCE for licence information 

;;  The framework where everything hangs together.
;; 

(in-package :cl-store)
(declaim (optimize (speed 3) (safety 0) (debug 0)))


(defvar *nuke-existing-classes* nil
  "Do we overwrite existing class definitions on restoration.")
(defvar *store-class-superclasses* t
  "Whether or not to store the superclasses of a stored class.")
(defvar *store-class-slots* t
  "Whether or not to serialize slots which are class allocated.")


(defvar *default-backend*)
(defvar *current-backend*)


;; conditions
;; Should these be the only errors that are thrown 
;; from store and restore?
(define-condition store-error ()
  ((format-string :accessor format-string :initarg :format-string :initform "Unknown")
   (format-args :accessor format-args :initarg :format-args :initform nil))
  (:report (lambda (condition stream)
             (apply #'format stream (format-string condition) 
                    (format-args condition))))
  (:documentation "Error thrown when storing an object fails."))

(define-condition restore-error ()
  ((format-string :accessor format-string :initarg :format-string :initform "Unknown")
   (format-args :accessor format-args :initarg :format-args :initform nil))
  (:report (lambda (condition stream)
             (apply #'format stream (format-string condition) 
                    (format-args condition))))
  (:documentation "Error thrown when restoring an object fails."))

(defun store-error (format-string &rest args)
  (error 'store-error :format-string format-string :format-args args))

(defun restore-error (format-string &rest args)
  (error 'restore-error :format-string format-string :format-args args))


;; entry points
(defun store-to-file (obj place backend)
  (let* ((backend-type (stream-type backend))
         (element-type (ecase backend-type
                         (character 'character)
                         (integer '(unsigned-byte 8)))))
    (with-open-file (s place :element-type element-type
                       :direction :output :if-exists :supersede)
      (store obj s backend))))

(defgeneric store (obj place &optional backend) 
  (:documentation "Entry Point for storing objects.")
  (:method ((obj t) (place stream) &optional (backend *default-backend*))
    "Store OBJ into Stream PLACE using backend BACKEND."
    (let ((*current-backend* backend))
      (backend-store obj place backend)))
  (:method ((obj t) (place string) &optional (backend *default-backend*))
    "Store OBJ into file designator PLACE using backend BACKEND."
    (store-to-file obj place backend))
  (:method ((obj t) (place pathname) &optional (backend *default-backend*))
    "Store OBJ into file designator PLACE using backend BACKEND."
    (store-to-file obj place backend)))


(defgeneric backend-store (obj place backend)
  (:argument-precedence-order backend place obj)
  (:documentation "Method wrapped by store, override this method for 
custom behaviour (see circularities.lisp).")
  (:method ((obj t) (place t) (backend t))
    "The default. Checks the streams element-type, stores the backend code
     and calls store-object."
    (check-stream-element-type place backend)
    (store-backend-code place backend)
    (store-object obj place backend)
    obj))



(defun check-stream-element-type (stream backend)
  "Ensure that the stream-element-type of STREAM is compatible with BACKEND."
  (let ((stream-type (stream-element-type stream))
        (backend-type (stream-type backend)))
    (unless (subtypep stream-type backend-type)
      (store-error  "Streams element type is ~A, backend expecting ~A."
                    stream-type backend-type))))

(defun store-backend-code (stream backend)
  "Store magic-number of BACKEND, when present, into STREAM."
  (let ((code (magic-number backend)))
    (when code
      (ecase (stream-type backend)
        (character (store-string-code code stream))
        (integer (store-32-byte code stream))))))



(defun store-object (obj stream &optional (backend *current-backend*))
  "Store OBJ into STREAM. Not meant to be overridden, 
   use backend-store-object instead"
  (backend-store-object obj stream backend))

(defgeneric backend-store-object (obj stream backend)
  (:documentation
   "Wrapped by store-object, override this to do custom storing 
   (see circularities.lisp for an example).")
  (:method ((obj t) (stream t) (backend t))
    "The default, just calls internal-store-object."
    (internal-store-object obj stream backend)))


(defgeneric internal-store-object (obj place backend)
  (:documentation "Method which is specialized by defstore-? macros.")
  (:method ((obj t) (place t) (backend t))
    "If call falls back here then OBJ cannot be serialized with BACKEND."
    (store-error "Cannot store objects of type ~A with backend ~(~A~)."
                 (type-of obj) (name backend))))

;; restoration
(defgeneric restore (place &optional backend)
  (:documentation 
   "Restore and object FROM PLACE using BACKEND. Not meant to be 
   overridden, use backend-restore instead")
  (:method (place &optional (backend *default-backend*))
    "Entry point for restoring objects (setfable)."
    (let ((*current-backend* backend))
      (backend-restore place backend))))
  
(defgeneric backend-restore (place backend)
  (:argument-precedence-order backend place)
  (:documentation "Wrapped by restore. Override this to do custom restoration")
  (:method ((place stream) (backend t))
    "Restore the object found in stream PLACE using backend BACKEND.
     Checks stream-element-type and magic-number and 
     invokes backend-restore-object"
    (check-stream-element-type place backend)
    (check-magic-number place backend)
    (backend-restore-object place backend))
  (:method ((place string) (backend t))
    "Restore the object found in file designator PLACE using backend BACKEND."
    (restore-from-file place backend))
  (:method ((place pathname) (backend t ))
    "Restore the object found in file designator PLACE using backend BACKEND."
    (restore-from-file place backend)))

(defun restore-from-file (place backend)
  (let* ((backend-type (stream-type backend))
         (element-type (ecase backend-type
                         (character 'character)
                         (integer '(unsigned-byte 8)))))
    (with-open-file (s place :element-type element-type :direction :input)
      (restore s backend))))
     
(defun (setf restore) (new-val place)
  (store new-val place))


(defun check-magic-number (stream backend)
  "Check to see if STREAM actually contains a stored object for BACKEND."
  (let ((magic-number (magic-number backend)))
    (when magic-number
      (let ((val (ecase (stream-type backend)
                   (integer (read-32-byte stream))
                   (character (retrieve-string-code stream)))))
        (unless (equal val magic-number)
          (restore-error "Stream does not contain a stored object for backend ~A."
                         (name backend)))))))

(defun lookup-reader (val readers)
  (gethash val readers))

(defgeneric get-next-reader (place backend)
  (:documentation 
   "Method which must be specialized for BACKEND to return 
   the next function to restore an object from PLACE.")
  (:method ((place t) (backend t))
    "The default, throw an error."
    (restore-error "get-next-reader must be specialized for backend ~(~A~)."
                   (name backend))))

(defgeneric find-function-for-type (place backend)
  (:documentation 
   "Return a function registered with defrestore-? which knows
   how to retrieve an object from PLACE, uses get-next-reader.")
  (:method (place backend)
    (let* ((val (get-next-reader place backend))
           (reader (lookup-reader val (restorer-funs backend))))
      (cond ((and val reader) reader)
            ((not val) 
             (restore-error "~A is not registered with backend ~(~A~)."
                            val (name backend)))
            ((not reader)
             (restore-error "No restorer defined for ~A in backend ~(~A~)."
                            val (name backend)))))))

;; Wrapper for backend-restore-object so we don't have to pass
;; a backend object around all the time
(defgeneric restore-object (place &optional backend)
  (:documentation
   "Restore the object in PLACE using BACKEND")
  (:method ((place t) &optional (backend *current-backend*))
    (backend-restore-object place backend)))

(defgeneric backend-restore-object (place backend)
  (:documentation
   "Find the next function to call with BACKEND and invoke it with PLACE.")
  (:method ((place t) (backend t))
    "The default"
    (funcall (the function (find-function-for-type place backend))
             place)))


;; EOF