;;; -*- Mode: LISP; Syntax: ANSI-Common-Lisp; Base: 10 -*-
;; See the file LICENCE for licence information.

(in-package :cl-store)

(declaim (optimize (speed 3) (safety 0) (debug 0)))

(export '(*xml-backend* xml-backend defstore-xml defrestore-xml princ-and-store
          princ-xml restore-first with-tag first-child second-child get-child
          add-xml-mapping))

(eval-when (:compile-toplevel :load-toplevel :execute)
  (defvar *xml-backend*
    (defbackend xml :stream-type 'char :extends resolving-backend)))


;; The xml backend does not use any type codes
;; we figure it out when we read the tag of each object
(defvar *xml-mapping* (make-hash-table :test #'equal))
(defun add-xml-mapping (name)
  (setf (gethash name *xml-mapping*)
        (intern name)))

(add-xml-mapping "REFERRER")
(add-xml-mapping "INTEGER")
(add-xml-mapping "SIMPLE-STRING")
(add-xml-mapping "FLOAT")
(add-xml-mapping "RATIO")
(add-xml-mapping "CHARACTER")
(add-xml-mapping "COMPLEX")
(add-xml-mapping "SYMBOL")
(add-xml-mapping "CONS")
(add-xml-mapping "PATHNAME")
(add-xml-mapping "HASH-TABLE")
(add-xml-mapping "STANDARD-OBJECT")
(add-xml-mapping "CONDITION")
(add-xml-mapping "STRUCTURE-OBJECT")
(add-xml-mapping "STANDARD-CLASS")
(add-xml-mapping "BUILT-IN-CLASS")
(add-xml-mapping "ARRAY")
(add-xml-mapping "SIMPLE-VECTOR")
(add-xml-mapping "PACKAGE")

(defmethod get-next-reader ((place list) (backend xml-backend))
  (gethash (car place) *xml-mapping*))

;; required methods and miscellaneous util functions
(defun princ-xml (tag value stream)
  (format stream "<~A>~A</~A>" tag value tag))

(defun princ-and-store (tag obj stream)
  (format stream "<~A>" tag)
  (store-object obj stream)
  (format stream "</~A>" tag))


(defmacro with-tag ((tag stream) &body body)
  `(progn 
    (format ,stream "<~A>" ,tag)
    ,@body
    (format ,stream "</~A>" ,tag)))
  
(defun first-child (elmt)
  (first (xmls:node-children elmt)))

(defun second-child (elmt)
  (second (xmls:node-children elmt)))

(defun get-child (name elmt)
  (or (assoc name (xmls:node-children elmt) :test #'equal)
      (error 'restore-error
             :datum "No child called ~A in xml ~a"
             :args (list name elmt))))

(defun get-attr (name elmt)
  (cadr (assoc name (xmls:node-attrs elmt) :test #'equal)))

(declaim (inline restore-first))
(defun restore-first (place)
  (restore-object (first-child place)))
  

;; override backend restore to parse the incoming stream
(defmethod backend-restore ((place stream) (backend xml-backend))
  (let ((*restore-counter* 0)
        (*need-to-fix* nil)
        (*restored-values* (make-hash-table)))
    (check-stream-element-type place backend)
    (let ((obj (backend-restore-object (xmls:parse place) backend)))
      (dolist (fn *need-to-fix*)
        (funcall (the function fn)))
      obj)))





;; referrer, Required for a resolving backend
(defmethod store-referrer (ref stream (backend xml-backend))
  (princ-xml "REFERRER" ref stream))

(defrestore-xml (referrer place)
  (make-referrer (parse-integer (third place))))




;; integer
(defstore-xml (obj integer stream)
  (princ-xml "INTEGER" obj stream))

(defrestore-xml (integer place) 
  (parse-integer (third place)))

;; simple-string
(defun xml-dump-simple-string (string place)
  (princ-xml "SIMPLE-STRING" string place))


(defstore-xml (obj string stream)
  (if (typep obj 'simple-standard-string)
      (xml-dump-simple-string obj stream)
      (xml-dump-array obj stream)))

(defrestore-xml (simple-string place)
  (third place))


;; float
(defstore-xml (obj float stream)
  (with-tag ("FLOAT" stream)
    (multiple-value-bind (signif exp sign) 
        (integer-decode-float obj)
      (princ-and-store "SIGNIFICAND" signif stream)
      (princ-and-store "EXPONENT" exp stream)
      (princ-and-store "SIGN" sign stream)
      (princ-and-store "TYPE" (float-type obj) stream))))

(defrestore-xml (float place)
  (float (* (* (restore-first (get-child "SIGNIFICAND" place))
               (* 1.0d0 (expt 2 (restore-first (get-child "EXPONENT" place)))))
            (restore-first (get-child "SIGN" place)))
         (get-float-type (restore-first (get-child "TYPE" place)))))

;; ratio
(defstore-xml (obj ratio stream)
  (with-tag ("RATIO" stream)
    (princ-and-store  "NUMERATOR" (numerator obj) stream)
    (princ-and-store "DENOMINATOR" (denominator obj) stream)))

(defrestore-xml (ratio place)
  (/ (restore-first (get-child "NUMERATOR" place))
     (restore-first (get-child "DENOMINATOR" place))))

  
;; character
(defstore-xml (obj character stream)
  (princ-xml "CHARACTER" (char-code obj) stream))

(defrestore-xml (character place)
  (code-char (read-from-string (first-child place))))



;; complex
(defstore-xml (obj complex stream)
  (with-tag ("COMPLEX" stream)
    (princ-and-store "REALPART" (realpart obj) stream)
    (princ-and-store "IMAGPART" (imagpart obj) stream)))

(defrestore-xml (complex place)
  (complex (restore-first (get-child "REALPART" place))
           (restore-first (get-child "IMAGPART" place))))

;; symbol
(defstore-xml (obj symbol stream)
  (format stream "<SYMBOL>~A::~A</SYMBOL>"
          (package-name (symbol-package obj))
          (symbol-name obj)))

(defrestore-xml (symbol place)
  (read-from-string (first-child place)))


;; cons
(defstore-xml (obj cons stream)
  (with-tag ("CONS" stream)
    (with-tag ("CAR" stream)
      (store-object (car obj) stream))
    (with-tag ("CDR" stream)
      (store-object (cdr obj) stream))))


(defrestore-xml (cons place)
  (let ((ret (cons nil nil))
        (car (get-child "CAR" place))
        (cdr (get-child "CDR" place)))
    (resolving-object ret
      (setting car (restore-first car))
      (setting cdr (restore-first cdr)))))



;; pathname 
(defstore-xml (obj pathname stream)
  (with-tag ("PATHNAME" stream)
    (princ-and-store "DEVICE" (pathname-device obj) stream)
    (princ-and-store "DIRECTORY" (pathname-directory obj) stream)
    (princ-and-store "NAME" (pathname-name obj) stream)
    (princ-and-store "TYPE" (pathname-type obj) stream)
    (princ-and-store "VERSION" (pathname-version obj) stream)))

(defrestore-xml (pathname place)
  (make-pathname 
   :device (restore-first (get-child "DEVICE" place))
   :directory (restore-first (get-child "DIRECTORY" place))
   :name (restore-first (get-child "NAME" place))
   :type (restore-first (get-child "TYPE" place))
   :version (restore-first (get-child "VERSION" place))))


;; hash-table
(defstore-xml (obj hash-table stream)
  (with-tag ("HASH-TABLE" stream)
    (princ-and-store "REHASH-SIZE" (hash-table-rehash-size obj) stream)
    (princ-and-store "REHASH-THRESHOLD" (hash-table-rehash-threshold obj) 
                     stream)
    (princ-and-store "SIZE" (hash-table-size obj) stream)
    (princ-and-store "TEST"(hash-table-test obj) stream)
    (with-tag ("ENTRIES" stream)
      (loop for key being the hash-keys of obj
            for value being the hash-values of obj do
            (with-tag ("ENTRY" stream)
              (princ-and-store "KEY" key stream)
              (princ-and-store "VALUE" value stream))))))

(defrestore-xml (hash-table place)
  (let ((hash1 (make-hash-table 
                :rehash-size (restore-first (get-child "REHASH-SIZE" place))
                :rehash-threshold (restore-first 
                                   (get-child "REHASH-THRESHOLD" place))
                :size (restore-first (get-child "SIZE" place))
                :test (symbol-function (restore-first (get-child "TEST" place))))))
    (resolving-object hash1
      (dolist (entry (xmls:node-children (get-child "ENTRIES" place)))
        (let* ((key-place (first-child (first-child entry)))
               (val-place (first-child (second-child entry))))
          (setting-hash (restore-object key-place) 
                        (restore-object val-place)))))
    hash1))




;; objects, conditions and structures
(defun xml-dump-type-object (obj stream)
  (let* ((all-slots (remove-if-not (lambda (x)
                                     (slot-boundp obj (slot-definition-name x)))
                                   (compute-slots (class-of obj))))
         (slots (if *store-class-slots* 
                    all-slots
                    (remove-if #'(lambda (x) (eql (slot-definition-allocation x)
                                                  :class))
                               all-slots))))
    (with-tag ("SLOTS" stream)
      (dolist (slot slots)
        (with-tag ("SLOT" stream)
          (let ((slot-name (slot-definition-name slot)))
            (princ-and-store "NAME" slot-name stream)
            (princ-and-store "VALUE" (slot-value obj slot-name) stream)))))))

(defstore-xml (obj standard-object stream)
  (with-tag ("STANDARD-OBJECT" stream)
    (princ-and-store "CLASS" (type-of obj) stream)
    (xml-dump-type-object obj stream)))

(defstore-xml (obj condition stream)
  (with-tag ("CONDITION" stream)
    (princ-and-store "CLASS" (type-of obj) stream)
    (xml-dump-type-object obj stream)))


#+(or sbcl cmu)
(defstore-xml (obj structure-object stream)
  (with-tag ("STRUCTURE-OBJECT" stream)
    (princ-and-store "CLASS" (type-of obj) stream)
    (xml-dump-type-object obj stream)))

(defun restore-xml-type-object (place)
  (let* ((class (find-class (restore-first (get-child "CLASS" place))))
         (new-instance (allocate-instance class)))
    (resolving-object new-instance
      (dolist (slot (xmls:node-children (get-child "SLOTS" place)))
        (let ((slot-name (restore-first (get-child "NAME" slot))))
          (setting (slot-value slot-name) 
                   (restore-first (get-child "VALUE" slot))))))
    new-instance))

(defrestore-xml (standard-object place)
  (restore-xml-type-object place))

(defrestore-xml (condition place)
  (restore-xml-type-object place))

#+(or sbcl cmu)
(defrestore-xml (structure-object place)
  (restore-xml-type-object place))
         

;; classes
(defun store-slot (slot stream)
  (with-tag ("SLOT" stream)
    (princ-and-store "NAME" (slot-definition-name slot) stream)
    (princ-and-store "ALLOCATION" (slot-definition-allocation slot) stream)
    (princ-and-store "TYPE" (slot-definition-type slot) stream)
    (with-tag ("INITARGS" stream)
      (dolist (x (slot-definition-initargs slot))
        (princ-and-store "INITARG" x stream)))
    (with-tag ("READERS" stream)
      (dolist (x (slot-definition-readers slot))
        (princ-and-store "READER" x stream)))
    (with-tag ("WRITERS" stream)
      (dolist (x (slot-definition-writers slot))
        (princ-and-store "WRITER" x stream)))))

(defstore-xml (obj standard-class stream)
  (with-tag ("STANDARD-CLASS" stream)
    (princ-and-store "NAME" (class-name obj) stream)
    (with-tag ("SUPERCLASSES" stream)
      (loop for x in (class-direct-superclasses obj) do
            (unless (eql x (find-class 'standard-object))
              (princ-and-store "SUPERCLASS" 
                               (if *store-class-superclasses*
                                   x
                                   (class-name x))
                               stream))))
    (with-tag ("SLOTS" stream)
      (dolist (x (class-direct-slots obj))
        (store-slot x stream)))
    (princ-and-store "METACLASS" (type-of obj) stream)))



(defun xml-add-class (name slots superclasses metaclass)
  (ensure-class name :direct-slots slots
                :direct-superclasses superclasses
                :metaclass metaclass)
  #+clisp(add-methods-for-class name slots))

(defun get-values (values)
  (loop for value in (xmls:node-children values)
        collect (restore-first value)))

(defun get-slots (slots)
  (loop for slot in (xmls:node-children slots) 
        collect (list :name (restore-first (get-child "NAME" slot))
                      :allocation (restore-first (get-child "ALLOCATION" slot))
                      :type (restore-first (get-child "TYPE" slot))
                      :initargs (get-values (get-child "INITARGS" slot))
                      :readers (get-values (get-child "READERS" slot))
                      :writers (get-values (get-child "WRITERS" slot)))))

(defun get-superclasses (superclasses)
  (loop for superclass in (xmls:node-children superclasses)
        collect (restore-first superclass)))
        
(defrestore-xml (standard-class  place)
  (let* ((name (restore-first (get-child "NAME" place)))
         (superclasses (get-superclasses (get-child "SUPERCLASSES" place)))
         (slots (get-slots (get-child "SLOTS" place)))
         (metaclass (restore-first (get-child "METACLASS" place))))
    (cond (*nuke-existing-classes*
           (xml-add-class name slots superclasses metaclass))
          (t (aif (find-class name nil)
                  it
                  (xml-add-class name slots superclasses metaclass))))))

;; built-in-classes 
(defstore-xml (obj built-in-class stream)
  (princ-and-store "BUILT-IN-CLASS" (class-name obj) stream))

(defrestore-xml (built-in-class place)
  (find-class (restore-first place)))

;; I don't know if this really qualifies as a built-in-class but it 
;; does make things a bit easier
(defmethod internal-store-object ((obj (eql (find-class 'hash-table))) stream 
                                  (backend xml-backend))
  (princ-and-store "BUILT-IN-CLASS" 'cl:hash-table stream))


;; Arrays and vectors
(defstore-xml (obj array stream)
  (xml-dump-array obj stream))

(defun xml-dump-array (obj stream)
  (with-tag ("ARRAY" stream)
    (princ-and-store "DIMENSIONS" (array-dimensions obj) stream)
    (if (and (= (array-rank obj) 1)
             (array-has-fill-pointer-p obj))
        (princ-and-store "FILL-POINTER" (fill-pointer obj) stream)
        (princ-and-store "FILL-POINTER" nil stream))
    (princ-and-store "ELEMENT-TYPE" (array-element-type obj) stream)
    (multiple-value-bind (to offset) (array-displacement obj)
      (princ-and-store "DISPLACED-TO" to stream)
      (princ-and-store "DISPLACED-OFFSET" offset stream))
    (princ-and-store "ADJUSTABLE" (adjustable-array-p obj) stream)
    (with-tag ("VALUES" stream)
      (loop for x from 0 to (1- (array-total-size obj)) do
            (princ-and-store "VALUE" (row-major-aref obj x) stream)))))

(defrestore-xml (array place)
  (let* ((dimensions (restore-first (get-child "DIMENSIONS" place)))
         (fill-pointer (restore-first (get-child "FILL-POINTER" place)))
         (element-type (restore-first (get-child "ELEMENT-TYPE" place)))
         (displaced-to (restore-first (get-child "DISPLACED-TO" place)))
         (displaced-offset (restore-first (get-child "DISPLACED-OFFSET"
                                                     place)))
         (adjustable (restore-first (get-child "ADJUSTABLE" place)))
         (res (make-array dimensions  
                          :element-type element-type
                          :adjustable adjustable
                          :fill-pointer fill-pointer)))
    (when displaced-to
      (adjust-array res dimensions :displaced-to displaced-to
                    :displaced-index-offset displaced-offset))
    (resolving-object res
      (loop for value in (xmls:node-children (get-child "VALUES" place))
            for count from 0 do
            (let ((pos count))
              (setting (row-major-aref pos)
                       (restore-first value)))))))


#-clisp    
(defstore-xml (obj simple-vector stream)
  (with-tag ("SIMPLE-VECTOR" stream)
    (princ-and-store "LENGTH" (length obj) stream)
    (with-tag ("ELEMENTS" stream)
      (loop for x across obj do
            (princ-and-store "ELEMENT" x stream)))))

#-clisp
(defrestore-xml (simple-vector place)
  (let* ((size (restore-first (get-child "LENGTH" place)))
         (res (make-array size)))
    (resolving-object res
      (loop for element in (xmls:node-children (get-child "ELEMENTS" place))
            for index from 1 do
            (let ((copy (1- index)))
              (setting (aref copy)
                       (restore-first element)))))))
                                      


;; packages
(defstore-xml (obj package stream)
  (princ-and-store "PACKAGE" (package-name obj) stream))

(defrestore-xml (package place)
  (find-package (restore-first place)))

;; EOF